/*! @file
 * @brief
 * contains stuff related to statistical information about the progress of factorization
 */


//! data structures and objects related to statistical information about the progress of factorization
namespace statistical_data
{
 int Erfolgsaufteilung[7] = {0,0,0,0,0,0,0};
  /* fr statistische Auswertung, inwieweit die dynamische
     Factorbase fr das Finden der Relationen
     verantwortlich ist:
      [0]->gar nicht beteiligt,
      [1]-> mit einem Faktor beteiligt
      [2]-> mit zwei Faktoren beteiligt
      [3]-> mit drei Faktoren beteiligt
      [4]-> mit 4 Faktoren beteiligt
      [5]-> mit 5 Faktoren beteiligt
      [6]-> mit mehr als 5 Faktoren beteiligt
   */

 int directly_sieved_Dynamic_Factors = 0;
 int directly_sieved_Special_Factors = 0;

#ifndef IS_CLIENT
 // the following statistical data can only be collected during
 // a detailed analysis of the relations. Therefore these data
 // is restricted to servers and standanlone versions.
 int Special_hit = 0;
 int Special_to_dynamic_Factor_hit = 0;
 int gefundene_DLP_Zyklen = 0;
#endif

#ifdef IS_SERVER
 // only for a server: collect statistical data about clients

 class CClientStats : protected CmpqsFactortypes
 {
  private:
   static CMutex Mutex;
   static unsigned int overall_counter[Factortype_size];
   unsigned int counter[Factortype_size];
   time_t first_connect, timestamp;
  protected:
   static void lock() { Mutex.lock(); }
   static void unlock() { Mutex.unlock(); }
  public:

   CClientStats() : counter(), first_connect(time(NULL)), timestamp(0) { };
   ~CClientStats() { };

   void PutTimeStamp(const time_t t = time(NULL))
   {
     lock();
     timestamp=t;
     unlock();
   }

   void increment(const CmpqsFactortypes::Factortype FTyp)
   {
     lock();
     ++counter[FTyp]; ++overall_counter[FTyp];
     unlock();  
   }
  
   const string info() const
   {
     ostringstream os;
     char buf [32];
     lock();
     const double percentage =
      100.0*static_cast<double>(counter[static_prime]+counter[single_large_prime]+counter[double_large_prime])
       /static_cast<double>(overall_counter[static_prime]+overall_counter[single_large_prime]+overall_counter[double_large_prime]);
     ctime_r(&first_connect,buf); // get localtime string for first_connect
     //buf[24]='\0'; // get rid of trailing newline
     buf[16]='\0'; // cut seconds, year and trailing newline
     char* const datestr = &buf[4]; // be less verbose
     os << datestr << " - ";
     ctime_r(&timestamp,buf);
     buf[16]='\0'; // cut seconds, year and trailing newline
     os << datestr << ":"
        << setw(5) << counter[static_prime] << ","
        << setw(7) << counter[single_large_prime] << ","
        << setw(8) << counter[double_large_prime] << "|";
     os.setf(ios::right, ios::adjustfield);
     os.setf(ios::fixed, ios::floatfield);
     os << setw(5) << setprecision(1) << percentage << "%" << flush;
     unlock();
     return os.str();
   }
 };
 CMutex CClientStats::Mutex;
 unsigned int CClientStats::overall_counter[Factortype_size] = {0};

 typedef map<string,CClientStats> TAllClientStats;
 TAllClientStats AllClientStats; // this is a list of all clients ever connected to the server
 CMutex AllClientStats_Mutex;

#endif /* IS_SERVER */


 time_t StartZeitpunkt = 0; // Zeitpunkt, zu dem das Sieben gestartet wurde
 time_t PreviousZeitpunkt = 0; // Zeitpunkt der vorigen Meldung
 int StartFuellung; // Fllung des GLS bei Start des Aussiebens
 struct tsample { time_t Zeitpunkt; int Fuellung; };
 const int samples=1<<4;
 tsample sample[samples]; int i_newest_sample=0;

 void Restzeit_kalkulieren(void)
 {
   // eigentlich wird hier weniger die Restzeit abgeschtzt als
   // vielmehr die durchschnittlich bentigte Zeit zum Sieben einer
   // Relation ermittelt...
   // Fr eine Restzeitabschtzung mte man das Zeitverhalten eigentlich
   // extrapolieren (und zustzlich die Schwankungen zwischen den
   // "Zwischenankunftszeiten" der Relationen ausmitteln).
   // Dies wre aber ein hoffnungsloses Unterfangen, da die Systemlast
   // schwankt und (im Serverbetrieb) die Anzahl der siebenden Clients
   // unbekannt ist (und zudem auch die Fileclients die Ergebnisse
   // verflschen wrden).
   if (Fuellung_GLS==StartFuellung) return;
   const time_t now=time(NULL);
   const int i_oldest_sample=(i_newest_sample+1)%samples;

   /*
      now.x - old.x = delta.x  // relations
      now.y - old.y = delta.y  // time
      delta.y/delta.x = y per x  -> seconds per static relation

      wanted.x = missing relations = Primbasis_size - Fuellung_GLS; known
      wanted.y -> (estimated) time till arrival; to compute

      wanted.y = (delta.y/delta.x)*wanted.x
      ETA = estimated time of arrival = now.y + wanted.y

   */

   double t=difftime(now,sample[i_oldest_sample].Zeitpunkt);
   t/=static_cast<double>(Fuellung_GLS-sample[i_oldest_sample].Fuellung);
   double Restzeit=(Primbasis_size-Fuellung_GLS)*t;

   {
     // print estimated time of arrivel
     char buf[32]; struct tm my_tm;
     const time_t eta = now+static_cast<time_t>(Restzeit);
     localtime_r(&eta,&my_tm); asctime_r(&my_tm,buf);
     buf[24]='\0'; // get rid of trailing newline
     cout << "ETA: " << buf << ", ";
   }

   double days = floor(Restzeit/(60*60*24));
   int hours,mins,secs;
   Restzeit=fmod(Restzeit,60*60*24);
   secs=static_cast<int>(Restzeit); mins=secs/60; hours=mins/60;
   mins%=60; secs%=60;

   if (days)
    cout_status << "time left (estimated): " << days << "d " << hours << "h" << endl;
   else
    cout_status << "time left (estimated): " << hours << "h " << mins << "m "
         << secs << "s" << endl;

#if defined(VERBOSE)
   cout << "oldest sample: " << sample[i_oldest_sample].Zeitpunkt << ": " << sample[i_oldest_sample].Fuellung << endl;
   cout << "newest sample: " << sample[i_newest_sample].Zeitpunkt << ": " << sample[i_newest_sample].Fuellung << endl;
#endif

   // Ringbuffer nur dann aktualisieren, wenn
   // mind. 1 Minute vergangen und Fuellung sich seitdem erhht hat
   if (now-sample[i_newest_sample].Zeitpunkt>=60 && Fuellung_GLS>sample[i_newest_sample].Fuellung)
    {
      i_newest_sample=i_oldest_sample; // wegen Ringbuffer!!
      sample[i_newest_sample].Zeitpunkt=now;
      sample[i_newest_sample].Fuellung=Fuellung_GLS;
    }
 }

 void Statuslegende_ausgeben(void)
 {
   cout_status << "Legend of the status information:" << endl;
   cout_status << "#static relations{distribution}/#static factorbase," << endl
               << "#directly sieved dynamic factors (single large primes)/#dynamic-FB," << endl;
 #ifdef IS_CLIENT
   cout_status << "#special factors (double large primes) ->{distribution}" << endl;
 #else
   cout_status << "#special factors (double large primes)" << endl;
 #endif
 #ifndef USE_NETWORK
   // Sieboffset macht nur bei standalone-Version Sinn
   cout_status << " [sievepos]" << endl;
 #endif
   cout_status << " -> % progress" << endl;

   if (!StartZeitpunkt)
    {
     StartZeitpunkt=time(NULL); StartFuellung=Fuellung_GLS; // Initialisierung fr Restzeitberechnung
     PreviousZeitpunkt=StartZeitpunkt; // Zeitpunkt dieser Ausgabe merken
     for (int i=0; i<samples; ++i) { sample[i].Zeitpunkt=StartZeitpunkt; sample[i].Fuellung=Fuellung_GLS; }
    }
 }

 bool Statusmeldung(const bool force_now = false)
 {
   /* 
      Statusmeldung ausgeben, wenn seit voriger Statusmeldung/Statuslegende
      mindestens 10 Sekunden vergangen sind.
      Der Funktionswert ist true, falls eine Ausgabe erfolgte, ansonsten false
    */ 
   if (!force_now && difftime(time(NULL),PreviousZeitpunkt)<10.0) return false;
   // aktuellen Stand ausgeben

 #ifdef IS_SERVER
  {
    // print a sheet with all clients that ever have connected to this server
    AllClientStats_Mutex.lock();
    for (TAllClientStats::iterator pos=AllClientStats.begin(); pos!=AllClientStats.end(); ++pos)
      cout << (pos->first+"                    ").substr(0,20) << ":" << pos->second.info() << endl;
    AllClientStats_Mutex.unlock();
  }
 #endif

   // overall statistics
   cout_status << Fuellung_GLS
        << "{" << Erfolgsaufteilung[0]
        << "," << Erfolgsaufteilung[1]
        << "," << Erfolgsaufteilung[2]
        << "," << Erfolgsaufteilung[3]
        << "," << Erfolgsaufteilung[4]
        << "," << Erfolgsaufteilung[5]
        << "," << Erfolgsaufteilung[6]
        << "}/" << Primbasis_size << ","
        << directly_sieved_Dynamic_Factors << "/"
        << Dynamic_Factor_Relations.size() << ",";
 #ifdef IS_CLIENT
   cout << directly_sieved_Special_Factors;
 #else
   cout_status
        << Special_Factor_Relations.size()/2
           // BUG/remark: Special_Factor_Relations is not protected by a
           // mutex, we actually can get wrong results, if we read
           // Special_Factor_Relations.size() while another thread is
           // modifying the set. But since we do read-only access, this
           // should not cause any harm. (this actually happened once while
           // Zyklensuche was active)
        << "->{" << Special_to_dynamic_Factor_hit << "/" << Special_hit
        << "/" << gefundene_DLP_Zyklen << "}";
 #endif
 #ifndef USE_NETWORK
   // Sieboffset macht nur bei standalone-Version Sinn
   cout_status << " [" << Sieboffset << "]";
 #endif
   cout_status.setf(ios::left, ios::adjustfield);
   cout_status << " -> ca. " << setw(5) << setprecision(4)
        << 100.0*Fuellung_GLS/Primbasis_size
        << "%" << endl;
   Polynom.Daten_sichern(cout); // mpqs-Polynomdaten ausgeben
   PreviousZeitpunkt=time(NULL); // Zeitpunkt dieser Ausgabe merken
 #ifndef IS_CLIENT
   Restzeit_kalkulieren();
 #endif
   return true;
 }

} // end of namespace statistical_data
using statistical_data::Statusmeldung;
using statistical_data::Statuslegende_ausgeben;
